# Copyright 2021 Spotify AB


import logging

import pytest

from klio_core import config


@pytest.fixture
def caplog(caplog):
    """Set global test logging levels."""
    caplog.set_level(logging.DEBUG)
    return caplog


def _job_config_dict():
    return {
        "metrics": {"logger": {}, "timer_unit": "ns"},
        "allow_non_klio_messages": False,
        "number_of_retries": 3,
        "events": {
            "inputs": [
                {
                    "type": "pubsub",
                    "topic": "test-parent-job-out",
                    "subscription": "test-parent-job-out-sub",
                    "data_location": "gs://sigint-output/test-parent-job-out",
                }
            ],
            "outputs": [
                {
                    "type": "pubsub",
                    "topic": "test-job-out",
                    "data_location": "gs://sigint-output/test-job-out",
                }
            ],
        },
        "data": {
            "inputs": [
                {
                    "type": "gcs",
                    "location": "gs://sigint-output/test-parent-job-out",
                }
            ],
            "outputs": [
                {"type": "gcs", "location": "gs://sigint-output/test-job-out",}
            ],
        },
        "more": "config",
        "that": {"the": "user"},
        "might": ["include"],
    }


def _benchmark_config_dict():
    return {
        "dataset": "sigint.benchmarking_uris.episodes_random_20210822",
        "enabled": True,
        "run_id": "8c438095-587c-4abc-9f49-6a35ed5d8dc4",
        "job_id": None,
        "skip_deploy": False,
        "metadata_topic": None,
        "data_topic": None,
    }


@pytest.fixture
def job_config_dict():
    return _job_config_dict()


def _pipeline_config_dict():
    return {
        "project": "test-project",
        "staging_location": "gs://some/stage",
        "temp_location": "gs://some/temp",
        "worker_harness_container_image": "gcr.io/sigint/foo",
        "streaming": True,
        "experiments": ["beam_fn_api"],
        "region": "us-central1",
        "num_workers": 3,
        "max_num_workers": 5,
        "disk_size_gb": 50,
        "worker_machine_type": "n1-standard-4",
        "runner": "direct",
        "autoscaling_algorithm": "THROUGHPUT_BASED",
        "update": False,
    }


@pytest.fixture
def pipeline_config_dict():
    return _pipeline_config_dict()


def _config_dict():
    return {
        "job_config": _job_config_dict(),
        "pipeline_options": _pipeline_config_dict(),
        "job_name": "test-job",
        "version": 2,
    }


@pytest.fixture
def config_dict():
    return _config_dict()


def _klio_config():
    return config.KlioConfig(_config_dict())


def _klio_config_w_benchmark():
    kconfig = _config_dict()
    kconfig["job_config"]["benchmark"] = _benchmark_config_dict()
    return config.KlioConfig(kconfig)



# Copyright 2020 Spotify AB

import logging
import os

from klio.transforms import io as klio_io

from spotify_klio.transforms import clients


# define an I/O transform using the klio-specific avro source
# note: fast avro is default for py3 on beam
class _KlioReadFromHadesAvro(klio_io._KlioReadFromAvro):
    # TODO: This probably shouldn't override the whole method, instead create a
    # sub-method for this to extend
    # TODO: `show_unpublished` isn't yet supported - it should be added to
    # spotify-klio-core (once we setup a spotify-specific klio config)
    def _get_file_pattern(self, file_pattern, location, show_unpublished=True):
        # TODO: this should be a validator in klio_core.config
        if not any([file_pattern, location]):
            raise klio_io.KlioMissingConfiguration(
                "Must configure at least one of the following keys when "
                "reading from avro: `file_pattern`, `location`."
            )

        # TODO: this should be a validator in klio_core.config
        # TODO: also check for valid and invalid URI schemes
        if location and location.startswith("hades:///"):
            hades_client = clients.HadesClient(location)

            location = hades_client.get_gcs_uri(show_unpublished)
            if file_pattern is None:
                file_pattern = "part-*"  # default avro+hades
            # We don't normally log where we're reading from. But I figured
            # it could be helpful here, particularly when the exact URL
            # may not be known (how often is someone going to provide a
            # revision?). It's only logged once per instantiation.
            logging.getLogger("klio").info(
                "Reading avro files from the following GCS bucket: "
                "%s" % location
            )

        if all([file_pattern, location]):
            file_pattern = os.path.join(location, file_pattern)

        elif file_pattern is None:
            file_pattern = location

        return file_pattern


class KlioReadFromHadesAvro(klio_io.KlioReadFromAvro):
    """ReadFromAvro transform that also supports Hades Endpoints"""

    def __init__(self, *args, **kwargs):
        self._reader = _KlioReadFromHadesAvro(*args, **kwargs)
        self.__counter = klio_io._KlioIOCounter(
            "read", "KlioReadFromHadesAvro"
        )



# -*- coding: utf-8 -*-
# Copyright 2020 Spotify AB

__author__ = "sigint"
__version__ = "21.9.0.dev1"
__email__ = "sigint@spotify.com"
__description__ = "Internal plugin for klio"
__uri__ = "https://ghe.spotify.net/sigint/spotify-klio"



# Copyright 2021 Spotify AB

import collections
import datetime
import json
import logging
import time
import types

from google.cloud import pubsub_v1


BQ_DATE_FORMAT = "%Y-%m-%d %H:%M:%S.000000 UTC"

Result = collections.namedtuple(
    "Result",
    [
        "uri",
        "transform",
        "timestamp_started",
        "benchmark_run_id",
        "successful",
        "processed_time",
    ],
)


class KlioBenchmarkWrapper(object):
    """Wrap a function with benchmarking."""

    DEFAULT_EXC_MSG = (
        "Failed to benchmark function: {function}."
        "Last exception: {last_exception}"
    )

    DEFAULT_DATA_TOPIC = "projects/sigint/topics/user-benchmarking-job-data"

    def __init__(
        self, function, klio_context, exception=None, exception_message=None,
    ):
        self._function = function
        self._func_name = getattr(function, "__qualname__", function.__name__)
        self._klio_context = klio_context
        self._exception = exception or Exception
        self._exception_message = exception_message
        self._logger = logging.getLogger("klio")
        self._job_name = klio_context.config.job_name

        if not getattr(klio_context.config.job_config, "benchmark", False):
            raise RuntimeError(
                "Benchmarking decorator in use with missing configurations. "
                "Please provide `klio-job.yaml::job_config.benchmark` "
                "as a dictionary with the field `dataset` configured to the "
                "dataset you would like to benchmark against and the field "
                "`enabled` set to `True`. "
                "Else remove the `@benchmark` decorator."
            )
        elif not klio_context.config.job_config.benchmark["enabled"]:
            raise RuntimeError(
                "Benchmarking decorator in use but benchmarking is "
                "not enabled. Either (1) update the field "
                "`klio-job.yaml::job_config.benchmark.enabled` to True, "
                "(2) rerun you job with the flag `--benchmark` "
                "to enable benchmarking, or "
                f"(3) remove `@benchmark` decorator from {self._func_name} "
                "if you do not intend to benchmark this job."
            )
        self._benchmark_data_topic = (
            klio_context.config.job_config.benchmark["data_topic"]
            or self.DEFAULT_DATA_TOPIC
        )
        self._benchmark_run_id = klio_context.config.job_config.benchmark[
            "run_id"
        ]
        self._publisher_client = pubsub_v1.PublisherClient()

    def __call__(self, *args, **kwargs):

        incoming_item = args[1]
        successful, duration = False, 0
        entity_uri = incoming_item.element.decode("utf-8")
        try:
            timestamp_started = datetime.datetime.utcnow().strftime(
                BQ_DATE_FORMAT
            )
            start_time = time.monotonic()
            ret = self._function(*args, **kwargs)

            if isinstance(ret, types.GeneratorType):
                ret = next(ret)
            successful = True
            return ret

        except self._exception as e:
            self._raise_exception(e)

        finally:
            duration = time.monotonic() - start_time
            result = Result(
                uri=entity_uri,
                transform=self._func_name,
                timestamp_started=timestamp_started,
                benchmark_run_id=self._benchmark_run_id,
                successful=successful,
                processed_time=duration,
            )
            self.emit_benchmark(result)

    def _raise_exception(self, last_exception):
        # TODO: This is where we can capture error_message
        # for benchmark.
        if self._exception_message is None:
            self._exception_message = self.DEFAULT_EXC_MSG.format(
                function=self._func_name, last_exception=last_exception
            )
        self._logger.warning(self._exception_message)
        raise last_exception

    def _submit_callback(self, fut):
        try:
            fut.result()
        except Exception as e:
            msg = "Error emitting benchmark '{}': {}".format(fut.uri, e)
            self.logger.warning(msg)

    def emit_benchmark(self, result):
        result_json = {
            "uri": result.uri,
            "transform": result.transform,
            "benchmark_run_id": result.benchmark_run_id,
            "timestamp_started": result.timestamp_started,
            "successful": result.successful,
            "processed_time": result.processed_time,
        }
        self._logger.info(
            f"Emitting benchmark {result_json}"
            f"to topic {self._benchmark_data_topic}"
        )
        try:
            bytes_data = json.dumps(result_json).encode("utf-8")
            future = self._publisher_client.publish(
                self._benchmark_data_topic, bytes_data
            )
            future.uri = result.uri
            future.add_done_callback(self._submit_callback)
            return future
        except Exception as e:
            self._logger.error(
                f"Failed to write benchmark for {result.uri}: {e}"
            )
        else:
            self._logger.info(
                f"Successfully wrote benchmark for {result.uri}."
            )



import asyncio
from unittest import mock

import pytest

from klio.transforms import core
from klio.transforms import decorators
from klio_core.proto import klio_pb2

from spotify_klio.transforms import decorators as sp_decorators
from tests import conftest


# NOTE: When the config attribute is accessed (when setting up
# a metrics counter object), it will try to read a
# `/usr/src/config/.effective-klio-job.yaml` file.
patcher = mock.patch.object(
    core.RunConfig, "get", conftest._klio_config_w_benchmark
)
patcher.start()


@pytest.fixture
def kmsg():
    msg = klio_pb2.KlioMessage()
    msg.data.element = b"spotify:episode:45h7rggVBoRwJPKpRNFzsn"
    return msg


def test_benchmark(mocker, monkeypatch, kmsg):
    mock_function = mocker.Mock()
    mock_pubsub = mocker.Mock()
    mock_pubsub_client = mocker.Mock()
    future = asyncio.Future()
    mock_pubsub_client.publish.return_value = future
    mock_pubsub.PublisherClient.return_value = mock_pubsub_client
    monkeypatch.setattr(sp_decorators.kbenchmark, "pubsub_v1", mock_pubsub)

    @decorators._handle_klio
    @sp_decorators._benchmark
    def func(kmsg, *args, **kwargs):
        mock_function(*args, **kwargs)
        yield

    func(kmsg.SerializeToString())
    func(kmsg.SerializeToString())
    func(kmsg.SerializeToString())
    assert len(mock_function.mock_calls) == 3
    assert len(mock_pubsub_client.publish.mock_calls) == 3


def test_benchmark_raises_parents(kmsg, mocker, monkeypatch):
    with pytest.raises(RuntimeError):

        @sp_decorators.benchmark
        def func(kmsg, *args, **kwargs):
            raise RuntimeError

        func(conftest._klio_config, kmsg.data)


def test_benchmark_off(mocker, monkeypatch, kmsg):
    mock_function = mocker.Mock()
    mock_pubsub = mocker.Mock()
    mock_pubsub_client = mocker.Mock()
    mock_pubsub_client.publish.return_value = mocker.Mock()
    mock_pubsub.PublisherClient.return_value = mock_pubsub_client
    monkeypatch.setattr(sp_decorators.kbenchmark, "pubsub_v1", mock_pubsub)

    with mock.patch.object(core.RunConfig, "get", conftest._klio_config):
        with pytest.raises(RuntimeError):

            @decorators._handle_klio
            @sp_decorators._benchmark
            def func(kmsg, *args, **kwargs):
                mock_function(*args, **kwargs)
                yield

            func(kmsg.SerializeToString())



# Copyright 2021 Spotify AB

import functools

from klio.transforms import _utils as txf_utils
from klio.transforms import decorators as klio_decorators

from spotify_klio.transforms import _benchmark as kbenchmark


def _benchmark(*args, exception=None, exception_message=None, **kwargs):
    def inner(func_or_meth):
        with klio_decorators._klio_context() as ctx:
            kctx = ctx

        benchmark_wrapper = kbenchmark.KlioBenchmarkWrapper(
            function=func_or_meth,
            klio_context=kctx,
            exception=exception,
            exception_message=exception_message,
        )

        @functools.wraps(func_or_meth)
        def method_wrapper(self, incoming_item, *args, **kwargs):
            args = (self, incoming_item) + args
            return benchmark_wrapper(*args, **kwargs)

        @functools.wraps(func_or_meth)
        def func_wrapper(kctx, incoming_item, *args, **kwargs):
            args = (kctx, incoming_item) + args
            return benchmark_wrapper(*args, **kwargs)

        if klio_decorators.__is_method(func_or_meth):
            return method_wrapper
        return func_wrapper

    if args and callable(args[0]):
        return inner(args[0])

    return inner


@txf_utils.experimental()
def benchmark(*args, **kwargs):
    """Run the decorated method/function with a benchmark emitter
    in a separate thread.
    Publishes benchmark to topic
        projects/sigint/topics/user-benchmarking-job-data and
    Dataflow job https://console.cloud.google.com/dataflow/jobs/
        europe-west1/2021-08-23_11_55_08-11381037738354644031?project=sigint
        will write to bigquery table `sigint.user_benchmarks.job_data`

    Order Matters when used with other Klio decorators.
    When using the `@handle_klio`, then the
    `@benchmark` decorator should be applied to a method/function
    **after** `@handle_klio` .
    .. code-block:: python
        @handle_klio
        @benchmark
        def my_map_func(ctx, item):
            ctx.logger.info(f"Received {item.element}")

        class MyDoFn(beam.DoFn):
            @handle_klio
            @benchmark
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element}"
                )
    When using benchmarking with retries and timeout decorators,
    placement of the `@benchmark` decorator will affect benchmark results.
    If `@benchmark` is placed **before** `@timeout` or `@retry`, benchmark
    results will include results for messages that have timed out or have
    been retried.
    .. code-block:: python
        class MyDoFn(beam.DoFn):
            @handle_klio
            @timeout
            @retry
            @benchmark
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element}"
                )
    If `@benchmark` is placed **after** `@timeout` or `@retry`, benchmark
    results will only include results for messages that have completed the
    logic of the inner `@timeout` or `@retry` logic.
    .. code-block:: python
        class MyDoFn(beam.DoFn):
            @handle_klio
            @benchmark
            @timeout
            @retry
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element}"
                )
    """
    return _benchmark(*args, **kwargs,)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Spotify AB

import codecs
import os
import re

from setuptools import find_packages
from setuptools import setup


HERE = os.path.abspath(os.path.dirname(__file__))


#####
# Helper functions
#####
def read(*filenames, **kwargs):
    """
    Build an absolute path from ``*filenames``, and  return contents of
    resulting file.  Defaults to UTF-8 encoding.
    """
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for fl in filenames:
        with codecs.open(os.path.join(HERE, fl), "rb", encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def find_meta(meta):
    """Extract __*meta*__ from META_FILE."""
    re_str = r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta)
    meta_match = re.search(re_str, META_FILE, re.M)
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


#####
# Project-specific constants
#####
NAME = "spotify-klio"
PACKAGE_NAME = "spotify_klio"
PACKAGES = find_packages(where="src")
META_PATH = os.path.join("src", PACKAGE_NAME, "__init__.py")
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Natural Language :: English",
    "Operating System :: POSIX :: Linux",
    "Operating System :: MacOS :: MacOS X",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
]
META_FILE = read(META_PATH)
INSTALL_REQUIRES = [
    "klio>=21.8.0",
    "requests",
]
EXTRAS_REQUIRE = {
    "docs": ["mkdocs", "interrogate"],
    "tests": [
        "coverage",
        "pytest>=4.3.0",  # 4.3.0 dropped last use of `convert`
        "pytest-cov",
        "pytest-mock",
    ],
}
EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["docs"] + EXTRAS_REQUIRE["tests"] + ["bumpversion", "wheel"]
)
# support 3.6, 3.7, & 3.8, matching Beam's support
PYTHON_REQUIRES = ">=3.6, <3.9"

setup(
    name=NAME,
    version=find_meta("version"),
    dependency_links=["https://pypi.spotify.net/simple"],
    description=find_meta("description"),
    url=find_meta("uri"),
    author=find_meta("author"),
    author_email=find_meta("email"),
    maintainer=find_meta("author"),
    maintainer_email=find_meta("email"),
    packages=PACKAGES,
    package_dir={"": "src"},
    include_package_data=True,
    classifiers=CLASSIFIERS,
    zip_safe=False,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=PYTHON_REQUIRES,
)



# Copyright 2020 Spotify AB

import requests


class HadesRevisionNotFound(Exception):
    pass


class HadesClient(object):
    """Basic client for making Hades Requests"""

    HADES_HOST = "https://hades.spotify.net/api/v3/revisions"
    URL_TPL = HADES_HOST + "/{endpoint}/{partition}"

    def __init__(self, location):
        self.hades_uri = location

    def _parse_hades_uri(self):
        _, uri = self.hades_uri.split("hades:///")
        endpoint, partition, *revision = uri.split("/")
        if len(revision):
            # *revision is a list, but if a revision is given, then it's
            # only one string
            revision = revision[0]
        return endpoint, partition, revision

    def get_gcs_uri(self, show_unpublished=False):
        endpoint, partition, revision = self._parse_hades_uri()

        url = self.URL_TPL.format(endpoint=endpoint, partition=partition)
        if revision:
            url = "{url}/{revision}".format(url=url, revision=revision)

        if show_unpublished:
            url = "{url}?showUnpublished".format(url=url)

        resp = requests.get(url)
        resp.raise_for_status()

        data = resp.json()["data"]
        if revision:
            # `uri` should be in `data`, otherwise we get a 404 which
            # will be caught in `resp.raise_for_status()`
            return data["uri"]

        revisions = data.get("revisions")
        if not revisions:
            raise HadesRevisionNotFound(
                "No revisions found for partition '{}' of endpoint "
                "'{}'.".format(partition, endpoint)
            )
        latest_rev = max(revisions, key=lambda r: r["creationTime"])
        return latest_rev["uri"]



import pytest

from klio.transforms import io as klio_io

from spotify_klio.transforms import io


@pytest.fixture
def mock_hades(mocker, monkeypatch):
    return mocker.patch.object(io.clients, "HadesClient")


def test_readfromhadesavro(mocker, mock_hades):

    mock_hades_client = mock_hades.return_value
    mock_hades_client.get_gcs_uri.return_value = "gs://gcs/location"

    mock_init = mocker.patch.object(klio_io.beam.io.ReadFromAvro, "__init__")
    mock_init.return_value = None

    mocker.patch.object(io.klio_io, "_KlioFastAvroSource")

    io.KlioReadFromHadesAvro(location="hades:///foo/bar")

    expected_file_pattern = "gs://gcs/location/part-*"

    mock_init.assert_called_once_with(
        file_pattern=expected_file_pattern,
        min_bundle_size=0,
        use_fastavro=True,
        validate=True,
    )
